/*
 * FreeRTOS V202002.00
 * Copyright (C) 2020 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * http://aws.amazon.com/freertos
 * http://www.FreeRTOS.org
 */

/**
 * @file iot_dual_outlet.c
 * @brief Implements the usage of the Dual Outlet functionality.
 */

#include "iot_dual_outlet.h"


/**************************************************************************//**
 *
 * @Name:  _getConnectedNetworkForDemo()
 *
 * @Description:  Obtain the Network type being used for running the application.
 *
 * @param  pDemoContext Structure storing the network parameters
 *
 * @retval  Network type
 *****************************************************************************/
static uint32_t _getConnectedNetworkForDemo( demoContext_t * pDemoContext )
{
    uint32_t ret = ( AwsIotNetworkManager_GetConnectedNetworks() & pDemoContext->networkTypes );

    if( ( ret & AWSIOT_NETWORK_TYPE_WIFI ) == AWSIOT_NETWORK_TYPE_WIFI )
    {
        ret = AWSIOT_NETWORK_TYPE_WIFI;
    }
    else if( ( ret & AWSIOT_NETWORK_TYPE_BLE ) == AWSIOT_NETWORK_TYPE_BLE )
    {
        ret = AWSIOT_NETWORK_TYPE_BLE;
    }
    else if( ( ret & AWSIOT_NETWORK_TYPE_ETH ) == AWSIOT_NETWORK_TYPE_ETH )
    {
        ret = AWSIOT_NETWORK_TYPE_ETH;
    }
    else
    {
        ret = AWSIOT_NETWORK_TYPE_NONE;
    }

    return ret;
}


/**************************************************************************//**
 *
 * @Name:  obtain_time()
 *
 * @Description:  Obtain the Unix EPOCH time using NTP (Network Time Protocol)
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
static void obtain_time(void)
{
    /*
     * Variables to store the maximum retry attempts and, the
     * current retry attempt when updating the local time over NTP
     */
    const int maxRetry = 10;
    int retry = 0;

    // SNTP setup and initializations
    sntp_setoperatingmode( SNTP_OPMODE_POLL );
    sntp_setservername( 0, "pool.ntp.org" );
    sntp_set_time_sync_notification_cb( time_sync_notification_cb );
    sntp_set_sync_mode( SNTP_SYNC_MODE_SMOOTH );
    sntp_init();

    /*
     * Periodic retry attempts every second to update the local/ system time
     * from the NTP servers.
     */
    while( (sntp_get_sync_status() == SNTP_SYNC_STATUS_RESET) &&
           (++retry < maxRetry) )
    {
        ESP_LOGI( TAG, "Waiting for system time to be set... (%d/%d)",
                    retry, maxRetry );

        vTaskDelay( MILLI_SECONDS_IN_ONE_SECOND / portTICK_PERIOD_MS );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvAnomalyDetection()
 *
 * @Description:  This function manages the system exceptions/ fault events
 *                that are generated and performs the necessary safety steps
 *                to prevent any potential damage. These include:
 *                01) GFCI Fault
 *                02) Power Failure
 *                03) Over Temperature
 *                04) Invalid Calculated Current
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvAnomalyDetection( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvAnomalyDetection() started.");

    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    while(1)
    {
        ESP_LOGI(TAG, "INFO : thread 6 - pvAnomalyDetection() "
                "with a periodic interval of 6 seconds");

        // variable xLastWakeTime is updated internally
        vTaskDelayUntil( &xLastWakeTime, pdMS_TO_TICKS(6000) );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvBlinkLedTask()
 *
 * @Description:  Thread to blink a LED at the desired rate
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvBlinkLedTask( void * pvParameters )
{
    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    /*
     * Variable to store the GPIO Pin Logic Level read using the
     * gpio_get_level() API.
     */
    BaseType_t gpio_pin_level = 0;

    /*
     * When using the gpio_get_level() to obtain the logic state of the GPIO Pin,
     * the GPIO Pin Mode must be configured as Input and Output.
     *
     * @warning If the pad is not configured for input (or input and output),
     *          the returned value of gpio_get_level() is always 0.
     *
     * Set Pin direction
     */
    gpio_set_direction( GPIO_NUM_32, GPIO_MODE_INPUT_OUTPUT );

    while(1)
    {
        // Obtain and save the GPIO Pin level
        gpio_pin_level = gpio_get_level( GPIO_NUM_32 );

        // Toggle the state of the GPIO Pin to blink the LED
        if( GPIO_PIN_LEVEL_LOW == gpio_pin_level )
        {
           gpio_set_level( GPIO_NUM_32, GPIO_PIN_LEVEL_HIGH );
        }
        else
        {
           gpio_set_level( GPIO_NUM_32, GPIO_PIN_LEVEL_LOW );
        }

        /* Delay associated with the desired blink rate */
//        vTaskDelayUntil( &xLastWakeTime,
//                        pdMS_TO_TICKS(LED_BLINK_RATE_ONE_HZ) );

        /*
         * Delay associated with the desired blink rate. Comment the above delay
         * code and uncomment the below delay to generate a new binary image,
         * which is used to validate that the OTA update is successful.
         */
        vTaskDelayUntil( &xLastWakeTime,
                        pdMS_TO_TICKS(LED_BLINK_RATE_TWO_HZ) );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvCalculateInstantaneousCurrent()
 *
 * @Description:  This function periodically calculates the instantaneous
 *                current being drawn from the outlets for the ongoing
 *                charging sessions.
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvCalculateInstantaneousCurrent( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvCalculateInstantaneousCurrent() started.");

    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    while(1)
    {
        ESP_LOGI(TAG, "INFO : thread 4 - pvCalculateInstantaneousCurrent() "
                "with a periodic interval of 4 seconds");

        // variable xLastWakeTime is updated internally
        vTaskDelayUntil( &xLastWakeTime, pdMS_TO_TICKS(4000) );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvCheckDeviceTemperature()
 *
 * @Description:  A periodic function that measures the device temperature.
 *                The Over heating temperature value equal 71 Â°C as per the
 *                OpenEVSE standards. The device temperature is compared with
 *                this reference value to take appropriate corrective measures
 *                when the device temperature reaches or crosses this value.
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvCheckDeviceTemperature( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvCheckDeviceTemperature() started.");

    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    while(1)
    {
        ESP_LOGI(TAG, "INFO : thread 7 - pvCheckDeviceTemperature() "
                "with a periodic interval of 7 seconds");

        // variable xLastWakeTime is updated internally
        vTaskDelayUntil( &xLastWakeTime, pdMS_TO_TICKS(7000) );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvCleanupTask()
 *
 * @Description:  Clean up the common libraries and the MQTT library.
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvCleanupTask( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvCleanupTask() started.");

    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    while(1)
    {
        /*
         * TODO: Change logic to Event driven by leveraging the IoTMutex_xxx or
         * the IoTSemaphore_xxx APIs.
         */
        if( true == bCleanup )
        {
            AwsIotNetworkManager_DisableNetwork( configENABLED_NETWORKS );
            /* Remove network manager subscription */
            AwsIotNetworkManager_RemoveSubscription( subscription );
            IotSemaphore_Destroy( &demoNetworkSemaphore );
            IotSdk_Cleanup();

            /*
             * This flag is updated from the various Threads running in parallel.
             * TODO Use this flag at appropriate locations and set it to true for
             * initiating a clean-up operation.
             */
            bCleanup = false;
        }

        /*
         * variable xLastWakeTime is updated internally.
         * Periodically check for a clean-up operation.
         */
        vTaskDelayUntil( &xLastWakeTime,
                         pdMS_TO_TICKS( PERIODIC_INTERVAL_1_SECOND ) );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvEsp32AndEnergyMeterCommunication()
 *
 * @Description:  This function handles the communication between the ESP32
 *                module and the Energy Meter module over UART.
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvEsp32AndEnergyMeterCommunication( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvEsp32AndEnergyMeterCommunication() started.");

    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    while(1)
    {
        ESP_LOGI(TAG, "INFO : thread 2 - pvEsp32AndEnergyMeterCommunication() "
                "with a periodic interval of 400 milliseconds");

        // variable xLastWakeTime is updated internally
        vTaskDelayUntil( &xLastWakeTime,
                         pdMS_TO_TICKS(ENERGY_METER_COMMUNICATION_PERIODIC_INTERVAL_DEFAULT) );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvEsp32GetTimeOverSntp()
 *
 * @Description:  Connect to the NTP Servers to synchronize and update the
 *                local time on this device with the Unix Epoch time. The local
 *                time is then internally updated. Additionally, the local time
 *                is periodically re-synchronized with the the NTP Server. The
 *                time interval for periodic synchronization is configurable.
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvEsp32GetTimeOverSntp( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvEsp32GetTimeOverSntp() started.");

    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    // Variables to store the 'time' value
    time_t now;
    struct tm timeinfo;
    char strftime_buf[64];

    // Obtain the current time and, store in the formatted timeinfo structure
    time( &now );
    localtime_r( &now, &timeinfo );

    /*
     * Is time set? If not, tm_year will be (1970)
     * Above condition holds true every time at device Power ON. Therefore,
     * update the local time by connecting with the NTP Server over the Internet.
     */
    if( timeinfo.tm_year < (2000) )
    {
        ESP_LOGI( TAG, "Time is not set yet. "
                  "Attempting to get the current time over NTP." );
        obtain_time();
    }

    while(1)
    {
        // update 'now' variable with current time
        time( &now );

        /*
         * Update the timestamp with the current Unix Epoch time, to be sent
         * along with the Energy Meter reading to the Server.
         *
         * TODO The below statement needs to be moved inside the
         * pvEsp32AndEnergyMeterCommunication Thread, wherein the 'timestamp'
         * variable needs to be updated every time the 'kWh' is updated.
         */
        emReading.timestamp = now;

        // Display on the serial console, time in Unix Epoch format
        ESP_LOGI( TAG, "Current UNIX EPOCH Time is: %ld", now );

        // Display on the serial console, formatted GMT time
        localtime_r( &now, &timeinfo );
        strftime( strftime_buf, sizeof(strftime_buf), "%c", &timeinfo );
        ESP_LOGI( TAG, "Equivalent GMT Date Time is: %s", strftime_buf );

        // variable xLastWakeTime is updated internally
        vTaskDelayUntil( &xLastWakeTime,
                         pdMS_TO_TICKS( PERIODIC_INTERVAL_10_SECONDS ) );
    }
}


/**************************************************************************//**
 *
 * @Name:  pvMqttTask()
 *
 * @Description: This Thread runs the MQTT task for periodically sending (PUB)
 *               the readings to the Server and, receiving (SUB) the messages
 *               associated to the subscribed topics from the Server.
 *
 * @param  pvParameters Parameters related to the mqttDEMO_entryFUNCTION
 *
 * @retval  None
 *****************************************************************************/
void pvMqttTask( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvMqttTask() started. Delaying task...");

    /*
     * Delay this task immediately once the thread is created and, let the
     * OTA thread complete establishing the MQTT connection, which will then be
     * used by this thread.
     *
     * TODO Replace TaskDelay with a synchronization primitive or move the
     * MQTT connection creation logic outside the Thread creation, as suggested
     * in the below post.
     * https://forums.freertos.org/t/running-mqtt-demo-periodically-with-ota-update-demo-causes-firmware-to-restart/10611/8?u=dipen12
     */
    vTaskDelay( pdMS_TO_TICKS(DELAY_15_SECONDS) );
    ESP_LOGI(TAG, "INFO : Thread pvMqttTask() re-started.");

    // Variables initialized with the default values
    demoContext_t * pContext = ( demoContext_t * ) pvParameters;
    const IotNetworkInterface_t * pNetworkInterface = NULL;
    void * pConnectionParams = NULL, * pCredentials = NULL;
    int status = EXIT_SUCCESS;

    /*
     * Get the Connected Network. Returns one of the supported <Network types>
     * as defined in <iot_network_types.h>
     */
    demoConnectedNetwork = _getConnectedNetworkForDemo( pContext );

    /*
     * Obtain the other required parameters required for running the function
     * associated with the MQTT operation.
     * pContext->demoFunction = MqttBasedCloudCommunication
     */
    pNetworkInterface = AwsIotNetworkManager_GetNetworkInterface( demoConnectedNetwork );
    pConnectionParams = AwsIotNetworkManager_GetConnectionParams( demoConnectedNetwork );
    pCredentials = AwsIotNetworkManager_GetCredentials( demoConnectedNetwork );

    // Run the MqttBasedCloudCommunication() function
    status = pContext->demoFunction( true,
                                     clientcredentialIOT_THING_NAME,
                                     pConnectionParams,
                                     pCredentials,
                                     pNetworkInterface );

    if( EXIT_SUCCESS != status )
    {
        ESP_LOGE(TAG, "Running the MqttBasedCloudCommunication() function "
                "resulted in an error.");
    }
}


/**************************************************************************//**
 *
 * @Name:  pvOtaUpdateTask()
 *
 * @Description:  Obtain the Network type being used for running the application.
 *
 * @param  pvParameters Parameters related to the otaUpdate_entryFUNCTION
 *
 * @retval  None
 *****************************************************************************/
void pvOtaUpdateTask( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvOtaUpdateTask() started.");

    // Variables initialized with the default values
    demoContext_t * pContext = ( demoContext_t * ) pvParameters;
    const IotNetworkInterface_t * pNetworkInterface = NULL;
    void * pConnectionParams = NULL, * pCredentials = NULL;
    int status = EXIT_SUCCESS;

    /*
     * Get the Connected Network. Returns one of the supported <Network types>
     * as defined in <iot_network_types.h>
     */
    demoConnectedNetwork = _getConnectedNetworkForDemo( pContext );

    /*
     * Obtain the other required parameters required for running the function
     * associated with the OTA Update operation.
     * pContext->demoFunction = vStartOTAUpdateDemoTask
     */
    pNetworkInterface = AwsIotNetworkManager_GetNetworkInterface( demoConnectedNetwork );
    pConnectionParams = AwsIotNetworkManager_GetConnectionParams( demoConnectedNetwork );
    pCredentials = AwsIotNetworkManager_GetCredentials( demoConnectedNetwork );

    // Run the vStartOTAUpdateDemoTask() function
    status = pContext->demoFunction( true,
                                     clientcredentialIOT_THING_NAME,
                                     pConnectionParams,
                                     pCredentials,
                                     pNetworkInterface );

    if( EXIT_SUCCESS != status )
    {
        ESP_LOGE(TAG, "Running the MqttBasedCloudCommunication() function "
                "resulted in an error.");
    }
}


/**************************************************************************//**
 *
 * @Name:  pvUpdateLedStripIndicator()
 *
 * @Description:  This function updates the LED strip colour corresponding to
 *                the various system events, thereby providing a visual
 *                notification to the user.
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvUpdateLedStripIndicator( void * pvParameters )
{
    ESP_LOGI(TAG, "INFO : Thread pvUpdateLedStripIndicator() started.");

    TickType_t xLastWakeTime;

    // We do this only once
    xLastWakeTime = xTaskGetTickCount();

    while(1)
    {
        ESP_LOGI(TAG, "INFO : thread 5 - pvUpdateLedStripIndicator() "
                "with a periodic interval of 5 seconds");

        // variable xLastWakeTime is updated internally
        vTaskDelayUntil( &xLastWakeTime, pdMS_TO_TICKS(5000) );
    }
}


/**************************************************************************//**
 *
 * @Name:  RunDualOutletFn()
 *
 * @Description:  The main application function, which creates multiple threads
 *                for the Dual Outlet functionality. This function creates all
 *                the required threads and terminates itself.
 *
 * @param  awsIotMqttMode         Specify if this application is running with the
 *                                AWS IoT MQTT server. Set this to `false` if
 *                                using another MQTT server.
 * @param  pIdentifier            NULL-terminated MQTT client identifier.
 * @param  pNetworkServerInfo     Passed to the MQTT connect function when
 *                                establishing the MQTT connection.
 * @param  pNetworkCredentialInfo Passed to the MQTT connect function when
 *                                establishing the MQTT connection.
 * @param  pNetworkInterface      The network interface to use for this application.
 *
 * @retval  EXIT_SUCCESS (0) Return status if all the required application Threads
 *                           are created successfully.
 * @retval  EXIT_FAILURE (1) Return status if any of the required application
 *                           Thread creation fails.
 *****************************************************************************/
int RunDualOutletFn( bool awsIotMqttMode,
                     const char * pIdentifier,
                     void * pNetworkServerInfo,
                     void * pNetworkCredentialInfo,
                     const IotNetworkInterface_t * pNetworkInterface )
{
    // Actual return status after executing this function.
    int status = EXIT_SUCCESS;

    /*
     * Parameters with their values, to carry out the MQTT based communication
     * with the Server.
     * .demoFunction = mqttDEMO_entryFUNCTION = MqttBasedCloudCommunication
     */
    static demoContext_t mqttDemoContext =
    {
        .networkTypes                = democonfigNETWORK_TYPES,
        .demoFunction                = mqttDEMO_entryFUNCTION,
        .networkConnectedCallback    = DEMO_networkConnectedCallback,
        .networkDisconnectedCallback = DEMO_networkDisconnectedCallback
    };

    /*
     * Parameters with their values, to enable the OTA Update of the firmware.
     * .demoFunction = otaUpdate_entryFUNCTION = vStartOTAUpdateDemoTask
     */
    static demoContext_t otaUpdateDemoContext =
    {
        .networkTypes                = democonfigNETWORK_TYPES,
        .demoFunction                = otaUpdate_entryFUNCTION,
        .networkConnectedCallback    = DEMO_networkConnectedCallback,
        .networkDisconnectedCallback = DEMO_networkDisconnectedCallback
    };

    /*
     * Test code to pass all variables associated with a single Energy Meter
     * reading, as a group within a structure.
     * These variables will be updated within the
     * pvEsp32AndEnergyMeterCommunication Thread, which communicates with the
     * Energy Meter module.
     */
    double kWh = 0034114.00;

    // Timestamp is updated from within the pvEsp32GetTimeOverSntp Thread.
//    emReading.timestamp = (long int)now;
    emReading.rfidCardUid = 0x5104212e;
    emReading.outletId = 0;
    emReading.kWh = kWh;

    /*
     * Create a Thread for updating the local/ system time in addition to periodic
     * synchronization with the NTP Server over the network.
     */
    status = (int) Iot_CreateDetachedThread( pvEsp32GetTimeOverSntp,
                                             NULL,
                                             IOT_THREAD_DEFAULT_PRIORITY,
                                             IOT_THREAD_DEFAULT_STACK_SIZE );

    if( true != status )
    {
        //TODO Thread creation failed
        ESP_LOGE(TAG, "Thread creation (pvEsp32GetTimeOverSntp) failed.");
        status = EXIT_FAILURE;
    }

    /*
     * LED Blink Thread
     * This Thread is created for testing purpose and, may be removed in the
     * application code.
     */
    status = (int) Iot_CreateDetachedThread( pvBlinkLedTask,
                                             NULL,
                                             IOT_THREAD_DEFAULT_PRIORITY,
                                             IOT_THREAD_DEFAULT_STACK_SIZE );

    if( true != status )
    {
        //TODO Thread creation failed
        ESP_LOGE(TAG, "Thread creation (pvBlinkLedTask) failed.");
        status = EXIT_FAILURE;
    }

    /* OTA Update Thread */
    status = (int) Iot_CreateDetachedThread( pvOtaUpdateTask,
                                             &otaUpdateDemoContext,
                                             IOT_THREAD_DEFAULT_PRIORITY,
                                             IOT_THREAD_DEFAULT_STACK_SIZE );

    if( true != status )
    {
        //TODO Thread creation failed
        ESP_LOGE(TAG, "Thread creation (pvOtaUpdateTask) failed.");
        status = EXIT_FAILURE;
    }

    /* Mqtt Thread */
    status = (int) Iot_CreateDetachedThread( pvMqttTask,
                                             &mqttDemoContext,
                                             IOT_THREAD_DEFAULT_PRIORITY,
                                             IOT_THREAD_DEFAULT_STACK_SIZE );

    if( true != status )
    {
        //TODO Thread creation failed
        ESP_LOGE(TAG, "Thread creation (pvMqttTask) failed.");
        status = EXIT_FAILURE;
    }

    /*
     * Cleanup Thread.
     * This Thread needs to run before shutting down the device, for a smooth
     * device/ system shutdown.
     */
    status = (int) Iot_CreateDetachedThread( pvCleanupTask,
                                             NULL,
                                             IOT_THREAD_DEFAULT_PRIORITY,
                                             IOT_THREAD_DEFAULT_STACK_SIZE );

    if( true != status )
    {
        //TODO Thread creation failed
        ESP_LOGE(TAG, "Thread creation (pvCleanupTask) failed.");
        status = EXIT_FAILURE;
    }

    return status;
}


/**************************************************************************//**
 *
 * @Name:  time_sync_notification_cb()
 *
 * @Description:  Callback function to notify the application every time the
 *                local time is successfully synchronized with a SNTP server.
 *
 * @param  tv Structure storing the time value (updated current time)
 *
 * @retval  None
 *****************************************************************************/
void time_sync_notification_cb(struct timeval *tv)
{
    /*
     * Print a message on the serial console every time the local time is
     * successfully updated with the SNTP Server time.
     */
    ESP_LOGI(TAG, "Notification of a time synchronization event");
}


