/* The config header is always included first. */
#include "iot_config.h"

/* Header file inclusions for various dependencies */
#include "aws_demo.h"
#include "esp_err.h"
#include "esp_log.h"
#include "iot_demo_runner.h"
#include "iot_init.h"
#include "iot_network_manager_private.h"

/* Platform layer includes. */
#include "platform/iot_threads.h"

/* Dual Outlet device configuration related settings */
#include "iot_dual_outlet_config.h"

/* Header file containing the network types */
#include "types/iot_network_types.h"

/* Obtain and maintain Unix Epoch time locally, using the Simple NTP */
#include "time.h"
#include "lwip/apps/sntp.h"

// GPIO Logic levels
#define GPIO_PIN_LEVEL_LOW 0
#define GPIO_PIN_LEVEL_HIGH 1

/*
 * Extern functions and variables
 */
extern int MqttBasedCloudCommunication( bool awsIotMqttMode,
                                        const char * pIdentifier,
                                        void * pNetworkServerInfo,
                                        void * pNetworkCredentialInfo,
                                        const IotNetworkInterface_t * pNetworkInterface );

extern int vStartOTAUpdateDemoTask( bool awsIotMqttMode,
                                    const char * pIdentifier,
                                    void * pNetworkServerInfo,
                                    void * pNetworkCredentialInfo,
                                    const IotNetworkInterface_t * pNetworkInterface );

// Variables used within the pvCleanupTask Thread as part of the cleanup task.
extern IotSemaphore_t demoNetworkSemaphore;
extern IotNetworkManagerSubscription_t subscription;


/*
 * Static variable declarations
 */

/* Variable to run the pvCleanupTask Thread only when required. */
static bool bCleanup = false;

/*
 * The 'demoConnectedNetwork' is considered as NONE at the start and, updated
 * with the actual value at run time.
 */
static uint32_t demoConnectedNetwork = AWSIOT_NETWORK_TYPE_NONE;

/* TAG required for ESP_LOGI() log statements on the serial console. */
static const char *TAG = "dual_outlet";


/*
 * Forward declaration of network connected DEMO callback to be renamed from
 * #define in aws_demo_config.h
 */
#ifdef DEMO_networkConnectedCallback
    void DEMO_networkConnectedCallback( bool awsIotMqttMode,
                                        const char * pIdentifier,
                                        void * pNetworkServerInfo,
                                        void * pNetworkCredentialInfo,
                                        const IotNetworkInterface_t * pNetworkInterface );
#else
    #define DEMO_networkConnectedCallback    ( NULL )
#endif

/*
 * Forward declaration of network disconnected DEMO callback to be renamed from
 * #define in aws_demo_config.h
 */
#ifdef DEMO_networkDisconnectedCallback
    void DEMO_networkDisconnectedCallback( const IotNetworkInterface_t * pNetworkInterface );
#else
    #define DEMO_networkDisconnectedCallback    ( NULL )
#endif


/**************************************************************************//**
 *
 * @Name:  _getConnectedNetworkForDemo()
 *
 * @Description:  Obtain the Network type being used for running the application.
 *
 * @param  pDemoContext Structure storing the network parameters
 *
 * @retval  Network type
 *****************************************************************************/
static uint32_t _getConnectedNetworkForDemo( demoContext_t * pDemoContext );


/**************************************************************************//**
 *
 * @Name:  obtain_time()
 *
 * @Description:  Obtain the Unix EPOCH time using NTP (Network Time Protocol)
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
static void obtain_time(void);


/**************************************************************************//**
 *
 * @Name:  pvAnomalyDetection()
 *
 * @Description:  This function manages the system exceptions/ fault events
 *                that are generated and performs the necessary safety steps
 *                to prevent any potential damage. These include:
 *                01) GFCI Fault
 *                02) Power Failure
 *                03) Over Temperature
 *                04) Invalid Calculated Current
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvAnomalyDetection( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvBlinkLedTask()
 *
 * @Description:  Thread to blink a LED at the desired rate
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvBlinkLedTask( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvCalculateInstantaneousCurrent()
 *
 * @Description:  This function periodically calculates the instantaneous
 *                current being drawn from the outlets for the ongoing
 *                charging sessions.
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvCalculateInstantaneousCurrent( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvCheckDeviceTemperature()
 *
 * @Description:  A periodic function that measures the device temperature.
 *                The Over heating temperature value equal 71 Â°C as per the
 *                OpenEVSE standards. The device temperature is compared with
 *                this reference value to take appropriate corrective measures
 *                when the device temperature reaches or crosses this value.
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvCheckDeviceTemperature( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvCleanupTask()
 *
 * @Description:  Clean up the common libraries and the MQTT library.
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvCleanupTask( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvEsp32AndEnergyMeterCommunication()
 *
 * @Description:  This function handles the communication between the ESP32
 *                module and the Energy Meter module over UART.
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvEsp32AndEnergyMeterCommunication( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvEsp32GetTimeOverSntp()
 *
 * @Description:  Connect to the NTP Servers to synchronize and update the
 *                local time on this device with the Unix Epoch time. The local
 *                time is then internally updated. Additionally, the local time
 *                is periodically re-synchronized with the the NTP Server. The
 *                time interval for periodic synchronization is configurable.
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvEsp32GetTimeOverSntp( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvMqttTask()
 *
 * @Description: This Thread runs the MQTT task for periodically sending (PUB)
 *               the readings to the Server and, receiving (SUB) the messages
 *               associated to the subscribed topics from the Server.
 *
 * @param  pvParameters Parameters related to the mqttDEMO_entryFUNCTION
 *
 * @retval  None
 *****************************************************************************/
void pvMqttTask( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvOtaUpdateTask()
 *
 * @Description:  Obtain the Network type being used for running the application.
 *
 * @param  pvParameters Parameters related to the otaUpdate_entryFUNCTION
 *
 * @retval  None
 *****************************************************************************/
void pvOtaUpdateTask( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  pvUpdateLedStripIndicator()
 *
 * @Description:  This function updates the LED strip colour corresponding to
 *                the various system events, thereby providing a visual
 *                notification to the user.
 *
 *
 * @param  None
 *
 * @retval  None
 *****************************************************************************/
void pvUpdateLedStripIndicator( void * pvParameters );


/**************************************************************************//**
 *
 * @Name:  time_sync_notification_cb()
 *
 * @Description:  Callback function to notify the application every time the
 *                local time is successfully synchronized with a SNTP server.
 *
 * @param  tv Structure storing the time value (updated current time)
 *
 * @retval  None
 *****************************************************************************/
void time_sync_notification_cb(struct timeval *tv);

